from typing import Any, Optional

from eth_utils import to_canonical_address, to_checksum_address
from vyper.utils import keccak256


# TODO replace return type with upcoming AddressType wrapper
def get_create2_address(
    blueprint_bytecode: bytes, deployer_address: Any, salt: bytes
) -> str:
    _, _, initcode = parse_erc5202(blueprint_bytecode)

    initcode_hash = keccak256(initcode)

    prefix = b"\xFF"
    addr = to_canonical_address(deployer_address)
    if len(salt) != 32:
        raise ValueError(f"bad salt (must be bytes32): {salt!r}")

    create2_hash = keccak256(prefix + addr + salt + initcode_hash)

    return to_checksum_address(create2_hash[12:])


# basically copied from ERC5202 reference implementation
def parse_erc5202(blueprint_bytecode: bytes) -> tuple[int, Optional[bytes], bytes]:
    """
    Given bytecode as a sequence of bytes, parse the blueprint preamble and
    deconstruct the bytecode into:
        the ERC version, preamble data and initcode.
    Raises an exception if the bytecode is not a valid blueprint contract
    according to this ERC.
    arguments:
        blueprint_bytecode: a `bytes` object representing the blueprint bytecode
    returns:
        (version,
         None if <length encoding bits> is 0, otherwise the bytes of the data section,
         the bytes of the initcode,
        )
    """
    if blueprint_bytecode[:2] != b"\xFE\x71":
        raise ValueError("Not a blueprint!")

    erc_version = (blueprint_bytecode[2] & 0b11111100) >> 2

    n_length_bytes = blueprint_bytecode[2] & 0b11
    if n_length_bytes == 0b11:
        raise ValueError("Reserved bits are set")

    data_length = int.from_bytes(
        blueprint_bytecode[3 : 3 + n_length_bytes], byteorder="big"
    )

    if n_length_bytes == 0:
        preamble_data = None
    else:
        data_start = 3 + n_length_bytes
        preamble_data = blueprint_bytecode[data_start : data_start + data_length]

    initcode = blueprint_bytecode[3 + n_length_bytes + data_length :]

    if len(initcode) == 0:
        raise ValueError("Empty initcode!")

    return erc_version, preamble_data, initcode



from eth_utils import to_checksum_address
from vyper.semantics.types import (
    AddressT,
    BoolT,
    BytesM_T,
    BytesT,
    DArrayT,
    IntegerT,
    InterfaceT,
    SArrayT,
    StringT,
    StructT,
    TupleT,
)
from vyper.utils import unsigned_to_signed

from boa.util.abi import Address
from boa.vm.utils import ceil32, floor32


# wrap storage in something which looks like memory
class ByteAddressableStorage:
    def __init__(self, evm, address: Address, key: int):
        self.evm = evm
        self.address = address
        self.key = key

    def __getitem__(self, subscript):
        if isinstance(subscript, slice):
            ret = b""
            start = subscript.start or 0
            stop = subscript.stop
            i = self.key + start // 32
            while i < self.key + ceil32(stop) // 32:
                ret += self.evm.get_storage_slot(self.address, i)
                i += 1

            start_ofst = floor32(start)
            start -= start_ofst
            stop -= start_ofst
            return memoryview(ret[start:stop])
        else:  # pragma: no cover
            raise Exception("Must slice {self}")


class _Struct(dict):
    def __init__(self, name, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.struct_name = name

    def __repr__(self):
        return f"{self.struct_name}({super().__repr__()})"


def _get_length(mem, bound):
    ret = int.from_bytes(mem[:32], "big")
    if ret > bound:
        # perf -- this is an uninitialized variable; length and data are
        # both garbage. truncate length to the bound of the container of
        # this type. the data will still be garbage, but we avoid OOM.
        return bound
    return ret


def decode_vyper_object(mem, typ):
    if isinstance(typ, BytesM_T):
        # TODO tag return value like `vyper_object` does
        return mem[: typ.m].tobytes()
    if isinstance(typ, (AddressT, InterfaceT)):
        return to_checksum_address(mem[12:32].tobytes())
    if isinstance(typ, BoolT):
        return bool.from_bytes(mem[31:32], "big")
    if isinstance(typ, IntegerT):
        ret = int.from_bytes(mem[:32], "big")
        if typ.is_signed:
            return unsigned_to_signed(ret, 256)
        return ret
    if isinstance(typ, BytesT):
        length = _get_length(mem[:32], typ.length)
        return mem[32 : 32 + length].tobytes()
    if isinstance(typ, StringT):
        length = _get_length(mem[:32], typ.length)
        return mem[32 : 32 + length].tobytes().decode("utf-8")
    if isinstance(typ, SArrayT):
        length = typ.count
        n = typ.subtype.memory_bytes_required
        return [
            decode_vyper_object(mem[i * n : i * n + n], typ.subtype)
            for i in range(length)
        ]
    if isinstance(typ, DArrayT):
        length = _get_length(mem[:32], typ.length)
        n = typ.subtype.memory_bytes_required
        ofst = 32
        ret = []
        for _ in range(length):
            ret.append(decode_vyper_object(mem[ofst : ofst + n], typ.subtype))
            ofst += n
        return ret
    if isinstance(typ, StructT):
        ret = _Struct(typ.name)
        ofst = 0
        for k, subtype in typ.tuple_items():
            n = subtype.memory_bytes_required
            ret[k] = decode_vyper_object(mem[ofst : ofst + n], subtype)
            ofst += n
        return ret
    if isinstance(typ, TupleT):
        ret = []
        ofst = 0
        for _, subtype in typ.tuple_items():
            n = subtype.memory_bytes_required
            ret.append(decode_vyper_object(mem[ofst : ofst + n], subtype))
            ofst += n
        return tuple(ret)

    return f"unimplemented decoder for `{typ}`"  # pragma: no cover



import contextlib
from typing import Generator

import hypothesis
import pytest

import boa
from boa.profiling import get_call_profile_table, get_line_profile_table
from boa.vm.gas_meters import ProfilingGasMeter

# monkey patch HypothesisHandle. this fixes underlying isolation for
# hypothesis.given() and also hypothesis stateful functionality.
_old_init = hypothesis.core.HypothesisHandle.__init__


def _HypothesisHandle__init__(self, *args, **kwargs):
    _old_init(self, *args, **kwargs)

    t = self.inner_test

    def f(*args, **kwargs):
        with boa.env.anchor():
            t(*args, **kwargs)

    self.inner_test = f


hypothesis.core.HypothesisHandle.__init__ = _HypothesisHandle__init__  # type: ignore


def pytest_addoption(parser):
    parser.addoption(
        "--gas-profile",
        action="store_true",
        help="Profile gas used by contracts called in tests",
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "ignore_isolation: ignore test isolation")
    config.addinivalue_line("markers", "ignore_gas_profiling: do not report on gas")
    config.addinivalue_line("markers", "gas_profile: report on gas")


def pytest_collection_modifyitems(config, items):
    if config.getoption("gas_profile"):
        for item in items:
            ignore_gas_profiling = item.get_closest_marker("ignore_gas_profiling")
            if not ignore_gas_profiling:
                item.add_marker("gas_profile")


_task_list = set()
_stack = {}  # preserve the original order of tasks
_id = 0


def pytest_fixture_setup(fixturedef, request):
    if request.node.get_closest_marker("ignore_isolation"):
        return

    global _id
    task_id = _id
    _id += 1

    ctx = boa.env.anchor()
    _stack[task_id] = ctx

    ctx.__enter__()

    def finalize():
        _task_list.add(task_id)
        _work_task_list()

    fixturedef.addfinalizer(finalize)


def _work_task_list():
    while (task_id := next(reversed(_stack.keys()), None)) in _task_list:
        ctx = _stack.pop(task_id)
        _task_list.remove(task_id)
        ctx.__exit__(None, None, None)


@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_call(item: pytest.Item) -> Generator:
    ignore_isolation = item.get_closest_marker("ignore_isolation") is not None
    profiling_ignored = item.get_closest_marker("ignore_profiling") is not None
    profiling_enabled = item.get_closest_marker("gas_profile") is not None

    if profiling_enabled and profiling_ignored:
        raise ValueError("Cannot ignore profiling and profile at the same time")

    @contextlib.contextmanager
    def _toggle_profiling(enabled: bool = False) -> Generator:
        if enabled:
            with boa.env.gas_meter_class(ProfilingGasMeter):
                yield
        else:
            yield

    with _toggle_profiling(profiling_enabled):
        if not ignore_isolation:
            with boa.env.anchor():
                yield
        else:
            yield


def pytest_sessionfinish(session, exitstatus):
    if boa.env._cached_call_profiles:
        import sys

        from rich.console import Console

        console = Console(file=sys.stdout)
        console.print(get_call_profile_table(boa.env))
        console.print(get_line_profile_table(boa.env))



import os
import statistics
from dataclasses import dataclass
from functools import cached_property
from textwrap import dedent

from eth_utils import to_checksum_address
from rich.table import Table

from boa.contracts.vyper.ast_utils import get_fn_name_from_lineno, get_line
from boa.environment import Env


def _safe_relpath(path):
    try:
        return os.path.relpath(path)
    except ValueError:
        # on Windows, if path and curdir are on different drives, an exception
        # can be thrown
        return path


@dataclass(unsafe_hash=True)
class LineInfo:
    address: str
    contract_name: str
    lineno: int
    line_src: str
    fn_name: str


@dataclass(unsafe_hash=True)
class ContractMethodInfo:
    address: str
    contract_name: str
    fn_name: str


@dataclass
class Stats:
    count: int = 0
    avg_gas: int = 0
    median_gas: int = 0
    stdev_gas: int = 0
    min_gas: int = 0
    max_gas: int = 0

    def __init__(self, gas_data):
        self.count = len(gas_data)
        self.avg_gas = int(statistics.mean(gas_data))
        self.median_gas = int(statistics.median(gas_data))
        self.stdev_gas = int(statistics.stdev(gas_data) if self.count > 1 else 0)
        self.min_gas = min(gas_data)
        self.max_gas = max(gas_data)

    def get_str_repr(self):
        return iter(
            [
                str(self.count),
                str(self.avg_gas),
                str(self.median_gas),
                str(self.stdev_gas),
                str(self.min_gas),
                str(self.max_gas),
            ]
        )


class CallGasStats:
    def __init__(self):
        self.net_gas = []
        self.net_tot_gas = []

    def compute_stats(self, typ: str = "net_gas") -> None:
        gas_data = getattr(self, typ)
        setattr(self, typ + "_stats", Stats(gas_data))

    def merge_gas_data(self, net_gas: int, net_tot_gas: int) -> None:
        self.net_gas.append(net_gas)
        self.net_tot_gas.append(net_tot_gas)


@dataclass
class Datum:
    gas_used: int = 0
    gas_refunded: int = 0
    # child gas usage might be useful for
    child_gas_used: int = 0
    child_gas_refunded: int = 0

    @property
    def net_gas(self):
        return self.gas_used - self.gas_refunded

    @property
    def net_tot_gas(self):
        return (
            self.gas_used
            - self.child_gas_used
            - self.gas_refunded
            + self.child_gas_refunded
        )

    def merge(self, other):
        for s in self.__dict__:
            self.__dict__[s] += other.__dict__[s]

    def adjust_child(self, child_computation):
        # adjust a Datum for gas used by child computation
        # generally if it's a vyper contract, we adjust the gas (since
        # that will show up elsewhere in the profile), and if it's some
        # black box contract, back out the adjustment.
        self.gas_used -= child_computation.get_gas_used()
        self.gas_refunded -= child_computation.get_gas_refund()
        self.child_gas_used += child_computation.get_gas_used()
        self.child_gas_refunded += child_computation.get_gas_refund()


# profile for a single call
class _SingleComputation:
    def __init__(self, contract, computation):
        self.contract = contract
        self.computation = computation

    @cached_property
    def by_pc(self):
        ret = {}
        for pc, gas in self.computation._gas_meter._gas_used_of.items():
            ret[pc] = Datum(gas_used=gas)

        for pc, gas in self.computation._gas_meter._gas_refunded_of.items():
            ret.setdefault(pc, Datum()).merge(Datum(gas_refunded=gas))

        for pc, child in zip(self.computation._child_pcs, self.computation.children):
            ret[pc].adjust_child(child)

        for pc in self.computation.code._trace:
            # in py-evm, STOP, RETURN and REVERT do not call consume_gas.
            # so, we need to zero them manually.
            ret.setdefault(pc, Datum())

        return ret

    @cached_property
    def by_line(self):
        ret = {}
        line_map = self.contract.source_map["pc_pos_map"]
        current_line = None
        seen = set()
        for pc in self.computation.code._trace:
            if line_map.get(pc) is not None:
                current_line, _, _, _ = line_map[pc]

            if current_line is not None and pc not in seen:
                ret.setdefault(current_line, Datum())
                ret[current_line].merge(self.by_pc[pc])
                seen.add(pc)

        return ret


# line profile. mergeable across contracts
class LineProfile:
    def __init__(self):
        self.profile = {}

    @classmethod
    def from_single(cls, contract, computation):
        ret = cls()
        ret.merge_single(_SingleComputation(contract, computation))
        return ret

    def raw_summary(self):
        return list(self.profile.items())

    def merge_single(self, single: _SingleComputation) -> None:
        for line, datum in single.by_line.items():
            self.profile.setdefault((single.contract, line), Datum()).merge(datum)

    def merge(self, other: "LineProfile") -> None:
        for (contract, line), datum in other.profile.items():
            self.profile.setdefault((contract, line), Datum()).merge(datum)

    def summary(
        self, display_columns=("net_tot_gas",), sortkey="net_tot_gas", limit=10
    ):
        raw_summary = self.raw_summary()

        if sortkey is not None:
            raw_summary.sort(reverse=True, key=lambda x: getattr(x[1], sortkey))
        if limit is not None and limit > 0:
            raw_summary = raw_summary[:limit]

        tmp = []
        for (contract, line), datum in raw_summary:
            data = ", ".join(f"{c}: {getattr(datum, c)}" for c in display_columns)
            line_src = get_line(contract.compiler_data.source_code, line)
            x = f"{contract.address}:{contract.compiler_data.contract_name}:{line} {data}"
            tmp.append((x, line_src))

        just = max(len(t[0]) for t in tmp)
        ret = [f"{l.ljust(just)}  {r.strip()}" for (l, r) in tmp]
        return _String("\n".join(ret))

    def get_line_data(self):
        raw_summary = self.raw_summary()
        raw_summary.sort(reverse=True, key=lambda x: x[1].net_gas)

        line_gas_data = {}
        for (contract, line), datum in raw_summary:
            fn_name = get_fn_name_from_lineno(contract.ast_map, line)

            # here we use net_gas to include child computation costs:
            line_info = LineInfo(
                address=contract.address,
                contract_name=contract.compiler_data.contract_name,
                lineno=line,
                line_src=get_line(contract.compiler_data.source_code, line),
                fn_name=fn_name,
            )

            line_gas_data[line_info] = datum.net_gas

        return line_gas_data


# stupid class whose __str__ method doesn't escape (good for repl)
class _String(str):
    def __repr__(self):
        return self


# cache gas_used for all computation (including children)
def cache_gas_used_for_computation(contract, computation):
    profile = contract.line_profile(computation)
    env = contract.env
    contract_name = contract.compiler_data.contract_name

    # -------------------- CACHE CALL PROFILE --------------------
    # get gas used. We use Datum().net_gas here instead of Datum().net_tot_gas
    # because a call's profile includes children call costs.
    # There will be double counting, but that is by choice.

    sum_net_gas = sum([i.net_gas for i in profile.profile.values()])
    sum_net_tot_gas = sum([i.net_tot_gas for i in profile.profile.values()])

    fn = contract._get_fn_from_computation(computation)
    if fn is None:
        fn_name = "unnamed"
    else:
        fn_name = fn.name

    fn = ContractMethodInfo(
        contract_name=contract_name,
        address=to_checksum_address(contract.address),
        fn_name=fn_name,
    )

    env._cached_call_profiles.setdefault(fn, CallGasStats()).merge_gas_data(
        sum_net_gas, sum_net_tot_gas
    )

    s = env._profiled_contracts.setdefault(fn.address, [])
    if fn not in env._profiled_contracts[fn.address]:
        s.append(fn)

    # -------------------- CACHE LINE PROFILE --------------------
    line_profile = profile.get_line_data()

    for line, gas_used in line_profile.items():
        env._cached_line_profiles.setdefault(line, []).append(gas_used)

    # ------------------------- RECURSION -------------------------

    # recursion for child computations
    for _computation in computation.children:
        child_contract = env.lookup_contract(_computation.msg.code_address)

        # ignore black box contracts
        if child_contract is not None:
            cache_gas_used_for_computation(child_contract, _computation)


def _create_table(for_line_profile: bool = False) -> Table:
    table = Table(title="\n")

    table.add_column("Contract", justify="left", style="cyan", no_wrap=True, width=52)
    computation_column_width = 30
    if for_line_profile:
        computation_column_width = 79
    table.add_column(
        "Computation",
        justify="left",
        style="cyan",
        no_wrap=True,
        width=computation_column_width,
    )

    table.add_column("Count", style="magenta")
    table.add_column("Mean", style="magenta")
    table.add_column("Median", style="magenta")
    table.add_column("Stdev", style="magenta")
    table.add_column("Min", style="magenta")
    table.add_column("Max", style="magenta")

    return table


def get_call_profile_table(env: Env) -> Table:
    table = _create_table()

    cache = env._cached_call_profiles
    cached_contracts = env._profiled_contracts
    contract_vs_median_gas = []
    for profile in cache:
        cache[profile].compute_stats()
        contract_vs_median_gas.append(
            (cache[profile].net_gas_stats.median_gas, profile.address)
        )

    # arrange from most to least expensive contracts:
    sort_gas = sorted(contract_vs_median_gas, key=lambda x: x[0], reverse=True)
    sorted_addresses = list(set([x[1] for x in sort_gas]))

    # --------------- POPULATE TABLE -----------------

    for address in sorted_addresses:
        fn_vs_median_gas = []
        for profile in cached_contracts[address]:
            fn_vs_median_gas.append((cache[profile].net_gas_stats.median_gas, profile))

        # arrange from most to least expensive contracts:
        fn_vs_median_gas = sorted(fn_vs_median_gas, key=lambda x: x[0], reverse=True)

        for c, (_, profile) in enumerate(fn_vs_median_gas):
            stats = cache[profile]

            # only first line should be populated for name and address
            cname = ""
            fn_name = profile.fn_name
            stats = list(stats.net_gas_stats.get_str_repr())
            if c == 0:
                relpath = _safe_relpath(profile.contract_name)
                contract_data_str = (
                    f"Path: {os.path.dirname(relpath)}\n"
                    f"Name: {os.path.basename(relpath)}\n"
                    f"Address: {address}\n"
                )
                cname = contract_data_str
                fn_name = "\n\n\n" + profile.fn_name
                stats = [f"\n\n\n{value}" for value in stats]
            table.add_row(cname, fn_name, *stats)

        table.add_section()

    return table


def get_line_profile_table(env: Env) -> Table:
    contracts: dict = {}
    for lp, gas_data in env._cached_line_profiles.items():
        contract_uid = (lp.contract_name, lp.address)

        # add spaces so numbers take up equal space
        lineno = str(lp.lineno).rjust(3)
        gas_data = (lineno + ": " + dedent(lp.line_src), gas_data)

        contracts.setdefault(contract_uid, {}).setdefault(lp.fn_name, []).append(
            gas_data
        )

    table = _create_table(for_line_profile=True)
    for (contract_name, contract_address), fn_data in contracts.items():
        relpath = _safe_relpath(contract_name)
        contract_data_str = (
            f"Path: {os.path.dirname(relpath)}\n"
            f"Name: {os.path.basename(relpath)}\n"
            f"Address: {contract_address}\n"
            f"{'-'*52}"
        )

        table.add_row(
            contract_data_str,
            "\n\n\n" + "-" * 74,
            "\n\nCount\n-----",
            "\n\nMean\n-----",
            "\n\nMedian\n-----",
            "\n\nStdev\n-----",
            "\n\nMin\n-----",
            "\n\nMax\n-----",
        )

        num_fn = 0
        for fn_name, _data in fn_data.items():
            l_profile = []
            for code, gas_used in _data:
                if code.endswith("\n"):
                    code = code[:-1]
                stats = Stats(gas_used)
                data = (contract_name, fn_name, code, *stats.get_str_repr())
                l_profile.append(data)

            # sorted by median (x[5]):
            l_profile = sorted(l_profile, key=lambda x: int(x[5]), reverse=True)

            for c, profile in enumerate(l_profile):
                cname = ""
                if c == 0:
                    cname = f"Function: {fn_name}"
                table.add_row(cname, *profile[2:])

            if not num_fn + 1 == len(fn_data):
                table.add_row("-" * 52, "-" * 74, *["-----"] * (len(profile[2:]) - 1))
                num_fn += 1

        table.add_section()

    return table



"""
Handlers for the JupyterLab extension.
"""
import logging
from http import HTTPStatus
from multiprocessing.shared_memory import SharedMemory

from jupyter_server.base.handlers import APIHandler
from jupyter_server.serverapp import ServerApp
from jupyter_server.utils import url_path_join
from tornado.web import authenticated

from boa.integrations.jupyter.constants import NUL, PLUGIN_NAME, TOKEN_REGEX


class CallbackHandler(APIHandler):
    """
    Handler that receives a POST from jupyter.js when user interacts with their browser wallet.
    The token is used to identify the SharedMemory object to write the callback data to.
    Besides, the token needs to fulfill the expected regex that ensures it is a valid format.
    It expects the SharedMemory object has already been created via a BrowserSigner instance.
    """

    @authenticated  # ensure only authorized users can request the server
    def get(self, token: str):
        """Checks if a SharedMemory object with the given token exists."""
        try:
            memory = SharedMemory(token)
        except FileNotFoundError:
            self.set_status(HTTPStatus.NOT_FOUND)
            return self.finish({"error": f"Invalid token: {token}"})
        memory.close()
        self.set_status(HTTPStatus.NO_CONTENT)
        return self.finish()

    @authenticated  # ensure only authorized users can request the server
    def post(self, token: str):
        """Writes the request body to the SharedMemory object identified by the token."""
        body = self.request.body
        if not body:
            self.set_status(HTTPStatus.BAD_REQUEST)
            return self.finish({"error": "Request body is required"})

        try:
            memory = SharedMemory(token)
        except FileNotFoundError:
            self.set_status(HTTPStatus.NOT_FOUND)
            error = f"Invalid token: {token}"
            logging.warning(error)
            return self.finish({"error": error})

        try:
            memory.buf[: len(body) + 1] = body + NUL
        except ValueError:
            self.set_status(HTTPStatus.REQUEST_ENTITY_TOO_LARGE)
            max_len = memory.size - len(NUL)
            error = (
                f"Request body has {len(body)} bytes, but only {max_len} are allowed"
            )
            return self.finish({"error": error})
        finally:
            memory.close()

        self.set_status(HTTPStatus.NO_CONTENT)
        return self.finish()


def setup_handlers(server_app: ServerApp) -> None:
    """
    Register the handlers in the Jupyter server.
    :param server_app: The Jupyter server application.
    """
    web_app = server_app.web_app
    base_url = url_path_join(web_app.settings["base_url"], PLUGIN_NAME)
    web_app.add_handlers(
        host_pattern=".*$",
        host_handlers=[(rf"{base_url}/callback/({TOKEN_REGEX})$", CallbackHandler)],
    )
    server_app.log.info(f"Handlers registered in {base_url}")



import json
from time import sleep
from typing import Optional

import requests

SESSION = requests.Session()


def _fetch_etherscan(
    uri: str, api_key: Optional[str] = None, num_retries=10, backoff_ms=400, **params
) -> dict:
    """
    Fetch data from Etherscan API.
    Offers a simple caching mechanism to avoid redundant queries.
    Retries if rate limit is reached.
    :param uri: Etherscan API URI
    :param api_key: Etherscan API key
    :param num_retries: Number of retries
    :param backoff_ms: Backoff in milliseconds
    :param params: Additional query parameters
    :return: JSON response
    """
    if api_key is not None:
        params["apikey"] = api_key

    for _ in range(num_retries):
        res = SESSION.get(uri, params=params)
        res.raise_for_status()
        data = res.json()
        if data.get("result") != "Max rate limit reached":
            break
        sleep(backoff_ms / 1000)

    if int(data["status"]) != 1:
        raise ValueError(f"Failed to retrieve data from API: {data}")

    return data


def fetch_abi_from_etherscan(
    address: str, uri: str = "https://api.etherscan.io/api", api_key: str = None
):
    # resolve implementation address if `address` is a proxy contract
    address = _resolve_implementation_address(address, uri, api_key)

    # fetch ABI of `address`
    params = dict(module="contract", action="getabi", address=address)
    data = _fetch_etherscan(uri, api_key, **params)

    return json.loads(data["result"].strip())


# fetch the address of a contract; resolves at most one layer of indirection
# if the address is a proxy contract.
def _resolve_implementation_address(address: str, uri: str, api_key: Optional[str]):
    params = dict(module="contract", action="getsourcecode", address=address)
    data = _fetch_etherscan(uri, api_key, **params)
    source_data = data["result"][0]

    # check if the contract is a proxy
    if int(source_data["Proxy"]) == 1:
        return source_data["Implementation"]
    else:
        return address



from eth.db.account import AccountDB


def _touch_account_patcher(self, address):
    self._accessed_accounts.add(address)


_BOA_PATCHED = object()


def patch_pyevm_state_object(state_object):
    if getattr(state_object, "__boa_patched__", None) == _BOA_PATCHED:
        return
    accountdb = state_object._account_db
    accountdb._restore_touch_account = accountdb.touch_account
    accountdb.touch_account = _touch_account_patcher.__get__(accountdb, AccountDB)
    state_object.__boa_patched__ = _BOA_PATCHED


def unpatch_pyevm_state_object(state_object):
    if not getattr(state_object, "__boa_patched__", None) == _BOA_PATCHED:
        return
    accountdb = state_object._account_db
    accountdb.touch_account = accountdb._restore_touch_account
    delattr(accountdb, "_restore_touch_account")
    delattr(state_object, "__boa_patched__")



import contextlib
import textwrap

import vyper.ast as vy_ast
import vyper.semantics.analysis as analysis
from vyper.ast.utils import parse_to_ast
from vyper.codegen.core import anchor_opt_level
from vyper.codegen.function_definitions import generate_ir_for_function
from vyper.codegen.ir_node import IRnode
from vyper.evm.opcodes import anchor_evm_version
from vyper.exceptions import InvalidType
from vyper.ir import compile_ir, optimizer
from vyper.semantics.analysis.utils import get_exact_type_from_node

from boa.contracts.vyper.ir_executor import executor_from_ir

# id used internally for method id name
_METHOD_ID_VAR = "_calldata_method_id"


@contextlib.contextmanager
def anchor_compiler_settings(compiler_data):
    settings = compiler_data.settings
    with anchor_opt_level(settings.optimize), anchor_evm_version(settings.evm_version):
        yield


def compile_vyper_function(vyper_function, contract):
    """Compiles a vyper function and appends it to the top of the IR of a
    contract. This is useful for vyper `eval` and internal functions, where
    the runtime bytecode must be changed to add more runtime functionality
    (such as eval, and calling internal functions)
    (performance note: this function is very very slow!)
    """

    compiler_data = contract.compiler_data

    with anchor_compiler_settings(compiler_data):
        global_ctx = contract.global_ctx
        ifaces = compiler_data.interface_codes
        ast = parse_to_ast(vyper_function, ifaces)
        vy_ast.folding.fold(ast)

        # override namespace and add wrapper code at the top
        with contract.override_vyper_namespace():
            analysis.add_module_namespace(ast, ifaces)
            analysis.validate_functions(ast)

        ast = ast.body[0]
        func_t = ast._metadata["type"]

        external_func_info = generate_ir_for_function(ast, global_ctx, False)
        ir = external_func_info.common_ir

        entry_label = func_t._ir_info.external_function_base_entry_label

        ir = ["seq", ["goto", entry_label], ir]

        # use a dummy method id
        ir = ["with", _METHOD_ID_VAR, 0, ir]

        # first mush it with the rest of the IR in the contract to ensure
        # all labels are present, and then optimize all together
        # (use unoptimized IR, ir_executor can't handle optimized selector tables)
        _, contract_runtime = contract.unoptimized_ir
        ir = IRnode.from_list(["seq", ir, contract_runtime])
        ir = optimizer.optimize(ir)

        assembly = compile_ir.compile_to_assembly(ir)
        bytecode, source_map = compile_ir.assembly_to_evm(assembly)
        bytecode += contract.data_section
        typ = func_t.return_type

        # generate the IR executor
        ir_executor = executor_from_ir(ir, compiler_data)

        return ast, ir_executor, bytecode, source_map, typ


def generate_bytecode_for_internal_fn(fn):
    """Wraps internal fns with an external fn and generated bytecode"""

    contract = fn.contract
    fn_name = fn.func_t.name
    fn_args = ", ".join([arg.name for arg in fn.func_t.arguments])

    return_sig = ""
    fn_call = ""
    if fn.func_t.return_type:
        return_sig = f" -> {fn.func_t.return_type}"
        fn_call = "return "
    fn_call += f"self.{fn_name}({fn_args})"

    # same but with defaults, signatures, etc.:
    _fn_sig = []
    for arg in fn.func_t.arguments:
        sig_arg_text = f"{arg.name}: {arg.typ}"

        # check if arg has a default value:
        if arg.name in fn.func_t.default_values:
            default_value = fn.func_t.default_values[arg.name].value
            sig_arg_text += f" = {default_value}"

        _fn_sig.append(sig_arg_text)
    fn_sig = ", ".join(_fn_sig)

    wrapper_code = textwrap.dedent(
        f"""
        @external
        @payable
        def __boa_private_{fn_name}__({fn_sig}){return_sig}:
            {fn_call}
    """
    )
    return compile_vyper_function(wrapper_code, contract)


def generate_bytecode_for_arbitrary_stmt(source_code, contract):
    """Wraps arbitrary stmts with external fn and generates bytecode"""

    ast = parse_to_ast(source_code)
    vy_ast.folding.fold(ast)
    ast = ast.body[0]

    return_sig = ""
    debug_body = source_code

    if isinstance(ast, vy_ast.Expr):
        with contract.override_vyper_namespace():
            try:
                ast_typ = get_exact_type_from_node(ast.value)
                return_sig = f"-> {ast_typ}"
                debug_body = f"return {source_code}"
            except InvalidType:
                pass

    # wrap code in function so that we can easily generate code for it
    wrapper_code = textwrap.dedent(
        f"""
        @external
        @payable
        def __boa_debug__() {return_sig}:
            {debug_body}
    """
    )
    return compile_vyper_function(wrapper_code, contract)



from eth.vm.gas_meter import GasMeter


class NoGasMeter:
    """
    A trivial gas meter, which does not meter gas usage.
    Improves EVM runtime performance by about 10%.
    Useful if you don't care about gas usage but want a
    performance boost.
    """

    def __init__(self, start_gas, *args, **kwargs):
        pass

    def consume_gas(self, amount, reason):
        pass

    def refund_gas(self, amount):
        pass

    def return_gas(self, amount):
        pass


class ProfilingGasMeter(GasMeter):
    """
    A gas meter which tracks the gas usage of every single PC.
    Used for profiling (users can back out opcode, line or function
    gas usage).
    """

    def __init__(self, start_gas, *args, **kwargs):
        super().__init__(start_gas, *args, **kwargs)
        self._gas_used_of = {}  # mapping from PCs to gas used
        self._gas_refunded_of = {}  # mapping from PCs to gas refunded

    def _set_code(self, code):
        self._code = code

    @property
    def _pc(self):
        # at the time that gas is refunded, pc is = to real pc + 1
        # (due to implementation detail of py-evm CodeStream.)
        return self._code.program_counter - 1

    def consume_gas(self, amount: int, reason: str) -> None:
        super().consume_gas(amount, reason)
        self._gas_used_of.setdefault(self._pc, 0)
        self._gas_used_of[self._pc] += amount

    def return_gas(self, amount: int) -> None:
        super().return_gas(amount)
        self._gas_used_of.setdefault(self._pc, 0)
        self._gas_used_of[self._pc] -= amount

    def refund_gas(self, amount: int) -> None:
        super().refund_gas(amount)
        self._gas_refunded_of.setdefault(self._pc, 0)
        self._gas_refunded_of[self._pc] += amount



import json
import sys
import textwrap
from importlib.abc import MetaPathFinder
from importlib.machinery import SourceFileLoader
from importlib.util import spec_from_loader
from pathlib import Path
from typing import Any, Union

import vyper
from vyper.cli.vyper_compile import get_interface_codes
from vyper.compiler.phases import CompilerData

from boa.contracts.abi.abi_contract import ABIContractFactory
from boa.contracts.vyper.compiler_utils import anchor_compiler_settings
from boa.contracts.vyper.vyper_contract import (
    VyperBlueprint,
    VyperContract,
    VyperDeployer,
)
from boa.environment import Env
from boa.explorer import fetch_abi_from_etherscan
from boa.util.abi import Address
from boa.util.disk_cache import DiskCache

_Contract = Union[VyperContract, VyperBlueprint]


_disk_cache = None


def set_cache_dir(cache_dir="~/.cache/titanoboa"):
    global _disk_cache
    if cache_dir is None:
        _disk_cache = None
        return
    compiler_version = f"{vyper.__version__}.{vyper.__commit__}"
    _disk_cache = DiskCache(cache_dir, compiler_version)


def disable_cache():
    set_cache_dir(None)


set_cache_dir()  # enable caching, by default!


class BoaImporter(MetaPathFinder):
    def find_spec(self, fullname, path, target=None):
        path = Path(fullname.replace(".", "/")).with_suffix(".vy")

        for prefix in sys.path:
            to_try = Path(prefix) / path

            if to_try.exists():
                loader = BoaLoader(fullname, str(to_try))
                return spec_from_loader(fullname, loader)


class BoaLoader(SourceFileLoader):
    def get_code(self, fullname):
        # importlib docs say to return None, but that triggers an `ImportError`
        return ""

    def create_module(self, spec):
        ret = load_partial(self.path)

        # comply with PEP-302:
        ret.__name__ = spec.name
        ret.__file__ = self.path
        ret.__loader__ = self
        ret.__package__ = spec.name.rpartition(".")[0]
        return ret


sys.meta_path.append(BoaImporter())


def compiler_data(
    source_code: str, contract_name: str, deployer=None, **kwargs
) -> CompilerData:
    global _disk_cache
    if deployer is None:
        deployer = _get_default_deployer_class()

    def _ifaces():
        # use get_interface_codes to get the interface source dict
        # TODO revisit this once imports are cleaned up vyper-side
        ret = get_interface_codes(Path("."), {contract_name: source_code})
        return ret[contract_name]

    if _disk_cache is None:
        ifaces = _ifaces()
        return deployer.create_compiler_data(
            source_code, contract_name, interface_codes=ifaces, **kwargs
        )

    def func():
        ifaces = _ifaces()
        ret = deployer.create_compiler_data(
            source_code, contract_name, interface_codes=ifaces, **kwargs
        )
        with anchor_compiler_settings(ret):
            _ = ret.bytecode, ret.bytecode_runtime  # force compilation to happen
        return ret

    assert isinstance(deployer, type)
    deployer_id = repr(deployer)  # a unique str identifying the deployer class
    cache_key = str((contract_name, source_code, kwargs, deployer_id))
    return _disk_cache.caching_lookup(cache_key, func)


def load(filename: str | Path, *args, **kwargs) -> _Contract:  # type: ignore
    name = Path(filename).stem
    # TODO: investigate if we can just put name in the signature
    if "name" in kwargs:
        name = kwargs.pop("name")
    with open(filename) as f:
        return loads(f.read(), *args, name=name, **kwargs, filename=filename)


def loads(
    source_code,
    *args,
    as_blueprint=False,
    name=None,
    filename=None,
    compiler_args=None,
    **kwargs,
):
    d = loads_partial(source_code, name, filename=filename, compiler_args=compiler_args)
    if as_blueprint:
        return d.deploy_as_blueprint(**kwargs)
    else:
        return d.deploy(*args, **kwargs)


def load_abi(filename: str, *args, name: str = None, **kwargs) -> ABIContractFactory:
    if name is None:
        name = filename
    with open(filename) as fp:
        return loads_abi(fp.read(), *args, name=name, **kwargs)


def loads_abi(json_str: str, *args, name: str = None, **kwargs) -> ABIContractFactory:
    return ABIContractFactory.from_abi_dict(json.loads(json_str), name, *args, **kwargs)


def loads_partial(
    source_code: str,
    name: str = None,
    filename: str | Path | None = None,
    dedent: bool = True,
    compiler_args: dict = None,
) -> VyperDeployer:
    name = name or "VyperContract"  # TODO handle this upstream in CompilerData
    if dedent:
        source_code = textwrap.dedent(source_code)

    compiler_args = compiler_args or {}

    deployer_class = _get_default_deployer_class()
    data = compiler_data(source_code, name, deployer_class, **compiler_args)
    return deployer_class(data, filename=filename)


def load_partial(filename: str, compiler_args=None):
    with open(filename) as f:
        return loads_partial(
            f.read(), name=filename, filename=filename, compiler_args=compiler_args
        )


def from_etherscan(
    address: Any, name=None, uri="https://api.etherscan.io/api", api_key=None
):
    addr = Address(address)
    abi = fetch_abi_from_etherscan(addr, uri, api_key)
    return ABIContractFactory.from_abi_dict(abi, name=name).at(addr)


def _get_default_deployer_class():
    env = Env.get_singleton()
    if hasattr(env, "deployer_class"):
        return env.deployer_class
    return VyperDeployer


__all__ = []  # type: ignore



from typing import Any

from vyper.ast import parse_to_ast
from vyper.builtins._signatures import BuiltinFunction
from vyper.builtins.functions import DISPATCH_TABLE, STMT_DISPATCH_TABLE
from vyper.builtins.functions import abi_encode as abi_encode_ir
from vyper.builtins.functions import ir_tuple_from_args, process_inputs
from vyper.codegen.core import IRnode, needs_external_call_wrap
from vyper.evm.address_space import MEMORY
from vyper.semantics.analysis.base import VarInfo
from vyper.semantics.namespace import get_namespace
from vyper.semantics.types import TupleT
from vyper.semantics.types.function import ContractFunctionT
from vyper.utils import keccak256

from boa.util.abi import abi_decode, abi_encode
from boa.vm.py_evm import register_raw_precompile


class PrecompileBuiltin(BuiltinFunction):
    def __init__(self, name, args, return_type, address):
        # override BuiltinFunction attributes
        self._id = name
        self._inputs = args  # list[tuple[str, VyperType]]
        self._return_type = return_type

        # set the precompile address
        self._address = address

    @process_inputs
    def build_IR(self, expr, args, kwargs, context):
        # allocate buffer for data to pass to precompile
        ret_buf = context.new_internal_variable(self._return_type)

        # allocate args buffer
        args_as_tuple = ir_tuple_from_args(args)
        args_abi_t = args_as_tuple.typ.abi_type
        args_buf = context.new_internal_variable(args_as_tuple.typ)

        ret = ["seq"]

        # store abi-encoded argument at buf
        args_len = abi_encode_ir(
            args_buf, args_as_tuple, context, args_abi_t.size_bound(), returns_len=True
        )
        ret_len = self._return_type.abi_type.size_bound()

        addr = int.from_bytes(self._address, "big")

        # call precompile
        ret.append(["staticcall", "gas", addr, args_buf, args_len, ret_buf, ret_len])
        ret += [ret_buf]

        return IRnode.from_list(ret, typ=self._return_type, location=MEMORY)


# takes a user-provided signature and produces shim code for
# serializing and deserializing
# ex. precompile("def foo() -> uint256")
def precompile(user_signature: str, force: bool = False) -> Any:
    def decorator(func):
        vy_ast = parse_to_ast(user_signature + ": view").body[0]
        func_t = ContractFunctionT.from_FunctionDef(vy_ast, is_interface=True)

        args_t = TupleT(tuple(func_t.argument_types))

        def wrapper(computation):
            # Decode input arguments from message data
            msg_data = computation.msg.data_as_bytes
            arg_values = abi_decode(args_t.abi_type.selector_name(), msg_data)

            # Call the original function with decoded input arguments
            res = func(*arg_values)

            return_t = func_t.return_type
            if return_t is not None:
                # Encode the result to be ABI-compatible
                # wrap to make it a tuple if necessary
                if needs_external_call_wrap(return_t):
                    res = (res,)
                    return_t = TupleT((return_t,))

                ret_abi_t = return_t.abi_type.selector_name()
                computation.output = abi_encode(ret_abi_t, res)

                return computation

        address = keccak256(user_signature.encode("utf-8"))[:20]
        register_raw_precompile(address, wrapper, force=force)

        args = [(arg.name, arg.typ) for arg in func_t.arguments]
        fn_name = func_t.name
        builtin = PrecompileBuiltin(fn_name, args, func_t.return_type, address)

        # sketchy check to see which dispatch table it should go in
        # ideally upstream vyper should be refactored to deal with this
        if func_t.return_type is not None:
            DISPATCH_TABLE[fn_name] = builtin
        else:
            STMT_DISPATCH_TABLE[fn_name] = builtin

        # yuck. note to refactor on vyper side.
        get_namespace()[fn_name] = VarInfo(builtin)

        return wrapper

    return decorator



from eth.vm.memory import Memory

from boa.vm.utils import ceil32, to_bytes, to_int


# a py-evm eth.vm.Memory compatible implementation of memory.
# most memory is aligned. add a cache which avoids converting
# between bytes and ints where possible
class FastMem(Memory):
    __slots__ = ("mem_cache", "_bytes", "needs_writeback")

    def __init__(self):
        # XXX: check if this would be faster as dict?
        self.mem_cache = []  # cached words

        # words which are in the cache but have not been written
        # to the backing bytes
        self.needs_writeback = []

        super().__init__()

    _DIRTY = object()

    def extend(self, start_position, size_bytes):
        # i.e. ceil32(len(self)) // 32
        new_size = (start_position + size_bytes + 31) // 32
        if (size_difference := new_size - len(self.mem_cache)) > 0:
            self.mem_cache.extend([self._DIRTY] * size_difference)
            self.needs_writeback.extend([False] * size_difference)
            super().extend(start_position, size_bytes)

    def read_word(self, start_position):
        if start_position % 32 == 0:
            if (ret := self.mem_cache[start_position // 32]) is not self._DIRTY:
                return ret

        ret = to_int(self.read_bytes(start_position, 32))
        self.mem_cache[start_position // 32] = ret
        return ret

    def _writeback(self, start_position, size):
        start = start_position // 32
        end = ceil32(start_position + size) // 32
        for ix in range(start, end):
            if self.needs_writeback[ix]:
                word = self.mem_cache[ix]
                assert ix + 32 <= len(self._bytes)
                self._bytes[ix * 32 : ix * 32 + 32] = to_bytes(word)
                self.needs_writeback[ix] = False

    def read_bytes(self, start_position, size):
        self._writeback(start_position, size)
        return super().read_bytes(start_position, size)

    def read(self, start_position, size):
        self._writeback(start_position, size)
        return super().read(start_position, size)

    def write_word(self, start_position, int_val):
        if start_position % 32 == 0:
            self.mem_cache[start_position // 32] = int_val
            self.needs_writeback[start_position // 32] = True
        else:
            self.write(start_position, 32, to_bytes(int_val))

    def write(self, start_position, size, value):
        start = start_position // 32
        end = (start_position + size + 31) // 32

        # need to write back, in case this is not an aligned write.
        self._writeback(start_position, size)
        for i in range(start, end):
            self.mem_cache[i] = self._DIRTY
            assert self.needs_writeback[i] is False

        super().write(start_position, size, value)



from boa.integrations.jupyter.browser import BrowserEnv, BrowserRPC, BrowserSigner
from boa.integrations.jupyter.constants import PLUGIN_NAME
from boa.integrations.jupyter.handlers import setup_handlers


def load_jupyter_server_extension(server_app):
    """
    Registers the API handler to receive HTTP requests from the frontend extension.
    :param server_app: jupyterlab.labapp.LabApp JupyterLab application instance.
    """
    setup_handlers(server_app)
    server_app.log.info(f"Registered {PLUGIN_NAME} server extension")


# Reference the old function name with the new function name.
_load_jupyter_server_extension = load_jupyter_server_extension


__all__ = [  # type: ignore
    BrowserSigner,
    BrowserRPC,
    BrowserEnv,
    load_jupyter_server_extension,
    _load_jupyter_server_extension,
]



import pdb

from vyper.exceptions import VyperException


class BoaDebug(pdb.Pdb):
    prompt = "(boa-debug) "

    def __init__(self, computation, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.contract = computation.msg._contract
        self.contract._computation = computation

    @property
    def intro(self):
        computation = self.contract._computation
        ast_source = self.contract.find_source_of(computation)
        if ast_source is not None:
            return str(VyperException("breakpoint at:", ast_source))

    def start(self):
        self.set_trace()

    def do_show_contract(self, *args):
        print(self.contract, file=self.stdout)

    def do_show_frame(self, *args):
        print(self.contract.debug_frame(), file=self.stdout)



EIP1167_PREFIX = bytes.fromhex("363d3d373d3d3d363d73")
EIP1167_SUFFIX = bytes.fromhex("5af43d82803e903d91602b57fd5bf3")


def is_eip1167_contract(bytecode):
    return bytecode.startswith(EIP1167_PREFIX) and bytecode.endswith(EIP1167_SUFFIX)


def extract_eip1167_address(bytecode):
    assert is_eip1167_contract(bytecode)
    ret = bytecode.strip(EIP1167_PREFIX).strip(EIP1167_SUFFIX)
    assert len(ret) == 20
    return ret



from typing import Any, Callable, Iterable, Optional, Union

from eth_abi.grammar import BasicType, TupleType, parse
from eth_utils import to_checksum_address
from hypothesis import given
from hypothesis import strategies as st
from hypothesis.strategies import SearchStrategy
from hypothesis.strategies._internal.deferred import DeferredStrategy

import boa
from boa.contracts.vyper.vyper_contract import VyperFunction
from boa.environment import Env
from boa.util.abi import Address

# hypothesis fuzzing strategies, adapted from brownie 0.19.2 (86258c7bd)
# in the future these may be superseded by eth-stdlib.

TYPE_STR_TRANSLATIONS = {"byte": "bytes1", "decimal": "fixed168x10"}

ArrayLengthType = Union[int, list, None]
NumberType = Union[float, int, None]


# note: there are also utils in the vyper codebase we could use for
# this. also, in the future we may want to replace these with strategies
# that use vyper types instead of abi types.
def get_int_bounds(type_str: str) -> tuple[int, int]:
    """Returns the lower and upper bound for an integer type."""
    size = int(type_str.strip("uint") or 256)
    if size < 8 or size > 256 or size % 8:
        raise ValueError(f"Invalid type: {type_str}")
    if type_str.startswith("u"):
        return 0, 2**size - 1
    return -(2 ** (size - 1)), 2 ** (size - 1) - 1


class _DeferredStrategyRepr(DeferredStrategy):
    def __init__(self, fn: Callable, repr_target: str) -> None:
        super().__init__(fn)
        self._repr_target = repr_target

    def __repr__(self):
        return f"sampled_from({self._repr_target})"


def _exclude_filter(fn: Callable) -> Callable:
    def wrapper(*args: tuple, exclude: Any = None, **kwargs: int) -> SearchStrategy:
        strat = fn(*args, **kwargs)
        if exclude is None:
            return strat
        if callable(exclude):
            return strat.filter(exclude)
        if not isinstance(exclude, Iterable) or isinstance(exclude, str):
            exclude = (exclude,)
        strat = strat.filter(lambda k: k not in exclude)
        # make the filter repr more readable
        repr_ = strat.__repr__().rsplit(").filter", maxsplit=1)[0]
        strat._LazyStrategy__representation = f"{repr_}, exclude={exclude})"
        return strat

    return wrapper


def _check_numeric_bounds(
    type_str: str, min_value: NumberType, max_value: NumberType
) -> tuple[NumberType, NumberType]:
    lower, upper = get_int_bounds(type_str)
    min_final = lower if min_value is None else min_value
    max_final = upper if max_value is None else max_value
    if min_final < lower or max_final > upper or min_final > max_final:
        raise ValueError
    return min_final, max_final


@_exclude_filter
def _integer_strategy(
    type_str: str, min_value: Optional[int] = None, max_value: Optional[int] = None
) -> SearchStrategy:
    min_val, max_val = _check_numeric_bounds(type_str, min_value, max_value)
    return st.integers(min_val, max_val)


@_exclude_filter
def _decimal_strategy(
    min_value: NumberType = None, max_value: NumberType = None, places: int = 10
) -> SearchStrategy:
    min_value, max_value = _check_numeric_bounds("int128", min_value, max_value)
    return st.decimals(min_value=min_value, max_value=max_value, places=places)


def format_addr(t):
    if isinstance(t, str):
        t = t.encode("utf-8")
    return to_checksum_address(t.rjust(20, b"\x00"))


@_exclude_filter
def _address_strategy(env: Optional[Env] = None) -> SearchStrategy:
    if env is None:
        env = boa.env

    st_random_account = st.binary(min_size=20, max_size=20).map(Address)

    # TODO: add addresses generated by the env through the env.generate_address() method
    env_accounts = list({obj.address for obj in env._code_registry.values()})
    if len(env_accounts) > 0:
        st_env_account = st.sampled_from(env_accounts)
        st_account = st.one_of(st_env_account, st_random_account)
    else:
        st_account = st_random_account

    return _DeferredStrategyRepr(lambda: st_account, "accounts")


@_exclude_filter
def _bytes_strategy(
    abi_type: BasicType, min_size: Optional[int] = None, max_size: Optional[int] = None
) -> SearchStrategy:
    size = abi_type.sub
    if not size:
        return st.binary(min_size=min_size or 1, max_size=max_size or 64)
    if size < 1 or size > 32:
        raise ValueError(f"Invalid type: {abi_type.to_type_str()}")
    if min_size is not None or max_size is not None:
        raise TypeError("Cannot specify size for fixed length bytes strategy")
    return st.binary(min_size=size, max_size=size)


@_exclude_filter
def _string_strategy(min_size: int = 0, max_size: int = 64, **kwargs) -> SearchStrategy:
    return st.text(min_size=min_size, max_size=max_size, **kwargs)


def _get_array_length(var_str: str, length: ArrayLengthType, dynamic_len: int) -> int:
    if not isinstance(length, (list, int)):
        raise TypeError(
            f"{var_str} must be of type int or list, not '{type(length).__name__}''"
        )
    if not isinstance(length, list):
        return length
    if len(length) != dynamic_len:
        raise ValueError(
            f"Length of '{var_str}' must equal the number of dynamic "
            f"dimensions for the given array ({dynamic_len})"
        )
    return length.pop()


def _array_strategy(
    abi_type: BasicType,
    min_length: ArrayLengthType = 1,
    max_length: ArrayLengthType = 8,
    unique: bool = False,
    **kwargs: Any,
) -> SearchStrategy:
    if abi_type.arrlist[-1]:
        min_len = max_len = abi_type.arrlist[-1][0]
    else:
        dynamic_len = len([i for i in abi_type.arrlist if not i])
        min_len = _get_array_length("min_length", min_length, dynamic_len)
        max_len = _get_array_length("max_length", max_length, dynamic_len)
    if abi_type.item_type.is_array:
        kwargs.update(min_length=min_length, max_length=max_length, unique=unique)
    base_strategy = strategy(abi_type.item_type.to_type_str(), **kwargs)
    strat = st.lists(base_strategy, min_size=min_len, max_size=max_len, unique=unique)
    # swap 'size' for 'length' in the repr
    repr_ = "length".join(strat.__repr__().rsplit("size", maxsplit=2))
    strat._LazyStrategy__representation = repr_  # type: ignore
    return strat


def _tuple_strategy(abi_type: TupleType) -> SearchStrategy:
    strategies = [strategy(i.to_type_str()) for i in abi_type.components]
    return st.tuples(*strategies)


# XXX: maybe rename to `abi`
def strategy(type_str: str, **kwargs: Any) -> SearchStrategy:
    type_str = TYPE_STR_TRANSLATIONS.get(type_str, type_str)
    if type_str == "fixed168x10":
        return _decimal_strategy(**kwargs)
    if type_str == "address":
        return _address_strategy(**kwargs)
    if type_str == "bool":
        return st.booleans(**kwargs)  # type: ignore
    if type_str == "string":
        return _string_strategy(**kwargs)

    abi_type = parse(type_str)
    if abi_type.is_array:
        return _array_strategy(abi_type, **kwargs)
    if isinstance(abi_type, TupleType):
        return _tuple_strategy(abi_type, **kwargs)  # type: ignore

    base = abi_type.base
    if base in ("int", "uint"):
        return _integer_strategy(type_str, **kwargs)
    if base == "bytes":
        return _bytes_strategy(abi_type, **kwargs)

    raise ValueError(f"No strategy available for type: {type_str}")


# XXX: is this the right module for me?
def fuzz(fn: VyperFunction):
    # usage:
    # @boa.fuzz(contract.function)
    # def f(arg_a, arg_b, arg_c):
    #     ...
    # f()
    # TODO how to test default values for overloaded functions?
    # or maybe just leave that to the user.
    strategies = {
        arg.name: strategy(arg.typ.canonical_abi_type) for arg in fn.func_t.arguments
    }

    return given(**strategies)



import sys
import types


# take an exception instance, and strip frames in the target module
# from the traceback
def strip_internal_frames(exc, module_name=None):
    error_type, error, traceback = sys.exc_info()
    frame = traceback.tb_frame

    if module_name is None:
        # use the parent module of the module where the exception was raised
        module_name = frame.f_globals["__name__"].rsplit(".", 1)[0]

    # currently, module_name is always
    # `boa.contracts.vyper`, `boa.contracts.abi` or `boa.contracts`
    while frame.f_globals.get("__name__", "").startswith(module_name):
        frame = frame.f_back

    # kwargs incompatible with pypy here
    # tb_next=None, tb_frame=frame, tb_lasti=frame.f_lasti, tb_lineno=frame.f_lineno
    tb = types.TracebackType(None, frame, frame.f_lasti, frame.f_lineno)
    return error.with_traceback(tb)
